//***************************************************************************************************************/
// RF Power Meter 10MHz - 10GHz                                                                                */
// Designed by Joost Breed - August 2017                                                                       */
// Based on AD8317 Log Detector of Analog Devices                                                              */
// Datasheet: http://www.analog.com/media/en/technical-documentation/evaluation-documentation/AD8317.pdf       */
// V1.3   26-07-2017                                                                                           */
//   - Far more samples per second, 14500 instead of 3500.                                                     */
//   - Calibration function added. 20 calibration points can be added.                                         */
//   - Serial plotter of arduino IDE can be used to display measured power in time.                            */
// Contains profiles of two commonly used attenuators for FPV equipment. Mini-Circuits VAT-20W2+ and VAT-30W2+ */
// Attenuation can now be set with 1 decimal accuracy.                                                         */
// Note: Menu functions are in the Menu.ino file. It also has to be opened in the arduino IDE.                 */
//**************************************************************************************************************
// V1.8_i2c   07-03-2018
//  umgeschrieben auf I2C-Bus DD7LP und ins deutsche Übersetzt im  Januar 2018 von DD7LP. 
// Die Library LCD I2C findet man hier: 
// https://bitbucket.org/fmalpartida/new-liquidcrystal/downloads
// entfernen sie bitte alte Librarys aus dem Entwicklungsverzeichnis    
// I2C vom Display wie folgt anschließen : SDA gleich A4, SCL gleich A5 am Arduino    
// Die Hexadresse vom I2C Dsplay ist meist der Wert 0x27   wenn unbekannt die Routine I2C-Scanner verwenden    
//  
//  Von mir DL3ED auf Grafhik-Display OLED 1,3" 128x64 umgeschrieben.
//  Ebenso auf einen Adruino Micro Pro (Leonardo ETH) mit einem 32U4 CPU umprogrammiert.
//  Die Taster wurden auch nach Masse geändert um die Widerstände zu sparen.
//  letzte Änderung 08.02.2021
//
//*************************************************************************************************************************************************************/*
#include <EEPROM.h>  //Lirary für EEPROM-Nutzung,in Arduino enthalten
#include <Bounce2.h> // Tastenentprellroutine
#include <Adafruit_SH1106.h>                           

#define OLED_RESET 4

Adafruit_SH1106 display(OLED_RESET);   

int analogReadFast(byte ADCpin, byte prescalerBits=4);

//***************************************************************************************************************************************************************
// Tasten pins beim Arduino
const byte TASTE_AUF    = 4;
const byte TASTE_EING   = 5;
const byte TASTE_AB     = 6;
const byte TASTE_ENTPRELLEN_MS = 5;    // Button debounce time in ms.
//***************************************************************************************************************************************************************
// Analog/Digitalwandler
const byte AD_WANDL_PIN    = A0;       // Pin benutzt als A/D converter input, für die Leistungsmessung.
const float V_REF          = 2.048;    //  2.048 original  External reference voltage, der Bezugspunkt für die Messung. Am Poti 1,60 Volt einstellen. #######
const float AD_AUFLOESUNG  = 1023.0;   // AD Wandler Auflösung - 1 = 2^10 - 1.

const int SAMPLE_COUNT     = 15000;    // Anzahl der Samples, die für die Mittelwertbildung zu entnehmen sind, was zu einer Abtastzeit von ca. 500ms führt.
//***************************************************************************************************************************************************************
//*************************************** Strings in Progmem zur Erhaltung von SRAM
#define FS(x) (__FlashStringHelper*)(x) // Funktion zum Abrufen von Strings aus dem Flash-Speicher. String wird mit dem PROGMEM-Keyword in Flash gesetzt. um SRAM zu erhalten.
const char str_MHz[] PROGMEM = " MHz";
const char str_GHz[] PROGMEM = " GHz";
const char str_dB[] PROGMEM = " dB";
const char str_dBm[] PROGMEM = " dBm";
//***************************************************************************************************************************************************************
// Menus
const byte MENU_MESSUNGEN                   = 0;
const byte MENU_MIN_LEISTUNG                = 1;
const byte MENU_MAX_LEISTUNG                = 2;
const byte MENU_MODULATIONINDEX             = 3;
const byte MENU_FREQUENZ_SETZEN             = 4;
const byte MENU_ABSCHWAECHER_SETZEN         = 5;
const byte MENU_CALIBRIERUNG                = 6;
const byte MENU_CALIBRIERUNG_AUSWAHL_FREQ   = 7;
const byte MENU_CALIBRIERUNG_BEST_LOESCHEN  = 8;
const byte MENU_CALIBRIERUNG_SETZE_FREQ     = 9;
const byte MENU_CALIBRIERUNG_FRAGE_VERBUN   = 10;
const byte MENU_CALIBRIERUNG_EINGA_LEISTUNG = 11;
//*****************************************************************************************************************************************************************
// EEPROM Addresses
const short INIT_SCHEMA        = 0x0110;             // Schema, um zu überprüfen, ob das EEPROM zum ersten Mal initialisiert wurde.
const byte INIT_EEPROM_ADDRESS = 0;                  // Adresse, an die das Schema geschrieben werden soll.
const byte FREQ_EEPROM_ADDRESS = 2;                  // Ausgewählte Frequenz-Startadresse (int = 2 Byte).
const byte ATTENUATION_PROFILE_EEPROM_ADDRESS = 4;   // Ausgewähltes Abschwächungsprofil Startadresse (1 Byte).
const byte ATTENUATION_VALUE_EEPROM_ADDRESS   = 5;   // Ausgewählte Abschwächer-Startadresse (float = 4 Byte).
const byte CALIBRATION_DATA_EEPROM_ADDRESS    = 100; // Ausgewählte Kalibrierdaten Startadresse
//*****************************************************************************************************************************************************************
// EEPROM Standartwerte (Defaults)
const int FREQ_DEFAULT                 = 10;        // Standardmäßig gewählte Frequenz 10MHz.
const float ATTENUATION_DEFAULT        = 30.0;      // Standart gewählte Dämpfung 30dBm.
const byte ATTENUATION_PROFILE_DEFAULT = 0;         // Standardprofil (0=kundenspezifisch, 1=VAT-20W2+, 2=VAT-30W2+).

const byte MAX_ATTENUATION_DBM         = 70;        // Maximal mögliche Dämpfung in dBm.
const byte MAX_ATTENUATION_PROFILE     = 2;         // Aktuelle Einstellung des  Benutzerprofils.
const int MIN_FREQUENZ                 = 10;        // Minimale unterstützte Frequenz in MHz
const int MAX_FREQUENZ                 = 9990;      // Maximal unterstützte Frequenz in MHz.
//****************************************************************************************************************************************************************
//************************************************* Eine Frequenz kann vom Benutzer ausgewählt werden. 
//************************************************* Die gewählte Frequenz bestimmt die typischen Leistungsmerkmale aus dem Datenblatt.
//************************************************* Alle untenstehenden Werte sind den Seiten 3 und 4 des Datenblattes entnommen.
const int MIN_INPUT_DBM               = -57;          //********** Minimale Eingangsleistung. Wenn weniger, dann betrachten Sie als"nicht verbunden".
const int ABS_MAX_INPUT_DBM           = 5;            //********** Für das absolute Maximum nehmen wir +5 dBm. (Mehr als +12dBm zerstören das IC)
const int ANZ_CAL_FREQUENZEN          = 18;           //********** Anzahl der Frequenzen, von denen wir die Kalibrierpunkte kennen.
const int FREQUENZEN[ANZ_CAL_FREQUENZEN]           = { 10   , 20   , 50   , 100  , 200  , 300  , 400  , 500  , 600  , 700  , 800  , 900  , 1000 , 1800, 2200, 3600, 5800, 8000 };  // List of FREQUENZEN we know settings data of.
const float DBM_AT_0V[ANZ_CAL_FREQUENZEN]           = { 16.92, 16.92, 16.92, 16.74, 16.83, 16.65, 16.56, 16.22, 16.13, 15.77, 15.77, 15.50, 15.14, 14.6, 14.1, 11.4, 16.8, 21.8 };  // dBm value at 0V measured with a Marconi Instruments 2022E. FREQUENZEN higher than 2GHz according to the datasheet, extrapolated from lineair section.
const float VAT20W2_ATT_PROFILE[ANZ_CAL_FREQUENZEN] = { 19.9 , 20.1 , 20.1 , 20.1 , 20.0 , 20.0 , 20.0 , 20.0 , 20.0 , 20.0 , 20.0 , 20.0 , 19.8 , 19.7, 19.6, 18.9, 18.1, 18.0 };  // Attenuator profile of Minicircuits VAT-20W2+ attenuator. A value for each frequenz of the list.
const float VAT30W2_ATT_PROFILE[ANZ_CAL_FREQUENZEN] = { 30.4 , 30.4 , 30.4 , 30.3 , 30.4 , 30.6 , 30.6 , 30.4 , 29.9 , 30.5 , 30.4 , 30.6 , 30.2 , 31.1, 30.0, 29.4, 28.5, 28.5 };  // Attenuator profile of Minicircuits VAT-30W2+ attenuator. A value for each frequenz of the list.
                                                    //-60,  -55, -50, -45, -40, -35, -30, -25, -20, -15, -10,   -5,  0    dBm
const float ERROR_DBM[13]        = {  2,  0.5, 0.3, 0.2,   0,   0,   0,   0,   0,   0,   0,  0.5,  3 };  // Fehler der gemessenen Leistung von -60dBm bis 0dBm in 5dB Schritten. 

const float MV_DB_SLOPE     = 0.02233;       // Steilheit der AD8317-Kennlinie in -mV/dBm. Ist für alle Frequenzen gleich. 
// Bemerkung: Die DBM_AT_0V[] Werte basieren auf dem -10dBm-Wert aus dem Datenblatt. 
// d.h. wenn f=5.8GHz, dann notiert das Datenblatt 700mV bei -10dBm. Mit einer Steigung von -22mV/dBm ==> (700/22m) - 10 = 21,82dBm bei 0V.
// Leider können diese Werte bis zu 1.9GHz sehr stark variieren. In diesem Fall wird der typische Wert genommen.
// Ich hatte keine kalibrierte Stromquelle. Aber wer es hat, kann vielleicht eine Kalibrierfunktion in die Software einbauen und die Werte im EEPROM speichern.
//***************************************************************************************************************************************************************
// Variablen declaration
Bounce button_UP = Bounce();                // Entprellroutine für Aufwärtstaste
Bounce button_ENTER = Bounce();             // Entprellroutine für Enter Taste
Bounce button_DOWN = Bounce();              // Entprellroutine für Abwärtstaste
//******************************************   Gemessene und berechnete Werte  **********************************************************************************
const byte MAX_HISTORY_COUNT = 10;          // maximale Anzahl von Einträgen   20  ######################
unsigned long sum = 0;                      // Variable zur Berechnung der Summe der Messungen
float avg_volt;                             // Durchschnittliche Spannung
float avg_dbm;                              // Aktuelle mittlere Leistung in dBm
float pep_dbm;                              // Aktueller Spitzenwert der Hüllkurvenleistung in dBm
float custom_attenuation = 0;               // Anwenderpezifische Dämpfung in dBm. Wird verwendet, wenn kein Dämpfungsprofil ausgewählt ist.
float min_dbm_history[MAX_HISTORY_COUNT];   // Minimale Leistung Verlauf Zähler
float max_dbm_history[MAX_HISTORY_COUNT];   // Maximal Leistung Verlauf Zähler
float min_dbm = 1000;                       // Minimale Leistung basierend auf Verlauf
float max_dbm = -1000;                      // Maximale Leistung basierend auf Verlauf
float modulation_index;                     // Modulationsindex basierend auf Verlauf
float dbm_at_0V;                            // Leistung dbm bei 0V;
float attenuation;                          // Dämpfung für ausgewählte Frequenz
byte att_profile = 0;                       // Aktuell ausgewähltes Dämpfungsprofil (0=Anwenderspezifisch, 1=VAT-20W2+, 2=VAT-30W2+).
bool power_overload;                        // Ist dies der Fall, wird zu viel Leistung in das Gerät eingespeist.
bool power_notconnected;                    // Wenn wahr, wird keine oder zu wenig Leistung in das Gerät eingespeist.
int selected_freq;                          // Ausgewählte Frequenz in den Messungen.
int menu_freq;                              // Ausgewählte Frequenz im Menü.
int adc_min_wert;                           // Wird während der Messung verwendet, um den Minimalwert der Messungen zu verfolgen.
int adc_max_wert;                           // Wird während der Messung verwendet, um den maximalen Wert der Messungen zu verfolgen.
int gross = 0;                              // Große Schrift.
int watt = 0;                               // Watt Anzeige.
int dbm = 0;                                // dbm Ausgabe.
int uw = 0;                                 // uw Ausgeben.
int kallibrieren_nicht = 0;                 // Damit beim kallibrieren keine führenden Nullen kommen.
//******************************************************************************************************************************************************************
// Menu and Display
byte display_line, display_col; //*******************************************************Cursorposition auf lcd
byte current_menu;              //****************************************************** Aktuell ausgewähltes Menü wird angezeigt
//******************************************************************************************************************************************************************
// Calibrations Datenstruktur
typedef struct
{
  unsigned int frequenz;   //**************Frequenz
  float dBm_at_0V;          //********************************************************** dBm Wert at 0 Volt
} 
calibrationPoint;

const int MAX_CALIBRIERUNGS_PUNKTE = 20;                                                //*************** Maximale Anzahl der zu speichernden Kalibrierpunkte
calibrationPoint calibrationData[MAX_CALIBRIERUNGS_PUNKTE];                             //*************** Vom Anwender eingegebene Kalibrierdaten
calibrationPoint mergedCalibrationData[MAX_CALIBRIERUNGS_PUNKTE + ANZ_CAL_FREQUENZEN];  //*************** Die Kalibrierdaten aus dem DBM_AT_0V-Array wurden mit den Daten der Benutzerkalibrierung zusammengeführt.
int selected_calibration_point = 0;
float calibrated_power;                                                                 //**************** Gemessene Leistung während der Kalibrierung.
//*******************************************************************************************************************************************************************
//********************************************************************** SETUP ROUTINE  *****************************************************************************
void setup() 
{   
  display.begin(SH1106_SWITCHCAPVCC, 0x3C);                     // initialize with the I2C addr 0x3D (for the 128x64)
  display.SH1106_command(SH1106_SETCONTRAST);                   // Displayhelligkeit #####################################################
  display.SH1106_command(255);                                  // 0-256  
  
  analogReference(EXTERNAL);                                    //******************* ADC auf externe Referenz einstellen  
  initButtons();
  getSettingsFromEEPROM();
  mergeCalibrationData();
  displaySplashScreen();                                          //Starttext ######
  resetHistoryRingBuffers();
  attenuation = calculateAttenuation();
  dbm_at_0V = interpolatedDbmValue();  
  gross = 0;
}
// ******************************************************************************************************************************************************************
// hier beginnt das eigendliche Hauptprogramm in der Schleife
// Es ist so eingestellt ,das wir ca. 2 Messungen pro Sekunde haben
// Wenn eine höhere Geschindigkeit erforderlich ist bitte die Samplerate erhöhen
// ******************************************************************************************************************************************************************
void loop() 
{    
  takeSamples();            // ********************** Macht eine Messung(liest auch die Tasten während der Messaufnahme)
  calculate();              // ********************** Werte berechnen
  readButtons();            // ********************** Lese Eingabetasten

                            //*********************************************** Zeigt Ergebnisse oder Fehler
  if (checkInputBounds())   //********************* Überprüfen Sie, ob sich die Messwerte im zulässigen Bereich sind. Wenn true zurückgegeben wird , dann Überlauf!
  {
    if (current_menu == MENU_MESSUNGEN){   
      kallibrieren_nicht = 1; 
      displayMeasurements();                  // **********  Menue Messungen wird aufgerufen  *********************
      kallibrieren_nicht = 0;}
    else if (current_menu == MENU_MIN_LEISTUNG)
      displayMinPower();                      // **********  Menue Min. Leistung wird aufgerufen  ********************
    else if (current_menu == MENU_MAX_LEISTUNG)
      displayMaxPower();                      // **********  Menue Max. Leistung wird aufgerufen  ********************
    else if (current_menu == MENU_MODULATIONINDEX)
      displayPEPModulationIndex();            // **********  Menue Moduöation wird aufgerufen  ******************** 
  }
  else  
    displayOverloadWarning();     // **************** Display Überlauf Warnung.************************
 
  readButtons();       // *************************** Abfrage der Eingabetasten ***********************  
}
//******************************************************** Ende der Hauptprogrammschleife **********************************************************

// ******************************************************   Funktion hole Messungen  ****************************************************************
// Zählt die Anzahl der Messungen
// Das Sampling von SAMPLE_COUNT-Werten mit einigen Berechnungen dauert ca. 500ms..
// Wenn das Programm nach dem Spash-Bildschirm abstürzt, kann es sein, dass Sie einen Arduino mit einer anderen Taktrate als 16MHz haben.. 
// Ändern Sie den Vorteiler auf eine andere Nummer siehe: http://www.robotplatform.com/knowledge/ADC/adc_tutorial_2.html
// Oder verwenden Sie stattdessen einfach analogRead().

void takeSamples()
{
  int adc_value;             
  // **************************************** Reset min, max and sum
  adc_min_wert = 1000;
  adc_max_wert = -1000;
  sum = 0;  
  for (int i=0; i<SAMPLE_COUNT; i++)
  {    
    adc_value = analogReadFast(AD_WANDL_PIN, 4); // Eine Messung ~20us        
    // Ermitteln von Min, Max und Summe aus aktuellen Messdaten
    if (adc_value < adc_min_wert) adc_min_wert = adc_value;
    if (adc_value > adc_max_wert) adc_max_wert = adc_value;    
    sum += adc_value;
    if (i%2000==0) readButtons(); // Abfrage der Eingabetasten einmal alle 2000 Samples.    
  }      
}
//*********************************************************************************************************************************************************************
// * Diese Funktion liest den AD-Wandler inerhalb von 20 us anstelle der üblichen 116 us 
// * Die Funktion wurde von Albert entnommen, siehe: http://www.avdweb.nl/arduino/libraries/fast-10-bit-adc.html
// ********************************************************************************************************************************************************************/
int analogReadFast(byte ADCpin, byte prescalerBits=4)
{ 
  byte ADCSRAoriginal = ADCSRA;
  ADCSRA = (ADCSRA & B11111000) | prescalerBits;
  int adc = analogRead(ADCpin); 
  ADCSRA = ADCSRAoriginal;
  return adc;
}

//*************************************** Funktion, alle notwendigen Berechnungen durch **********************************************************************
void calculate()
{    
  float measured_min_dbm, measured_max_dbm;
  static byte history_idx = 0;                                // Aktueller Index der Verlaufs-Ringbuffer

  measured_min_dbm = measuredValueTodBm(adc_max_wert);       // Je niedriger die Spannung, desto höher die Leistung.
  measured_max_dbm = measuredValueTodBm(adc_min_wert);       // Wenn max_adc = min_dbm und min_adc = max_dbm.
  
  float avg_measurement = sum / SAMPLE_COUNT;                 // bildet den Mittelwert der Abtastungen
  avg_volt = avg_measurement * (V_REF / AD_AUFLOESUNG);       // Convertiert zu Spannung
  avg_dbm = measuredValueTodBm(avg_measurement);              // Convertiert Mittelwert zu dBm.  
  pep_dbm = measured_max_dbm;                                 // Die maximale Hüllkurvenleistung ist die maximal gemessene Leistung.
 
  // Speichern des aktuellen Min/Max-Wertes in den Verlaufsprotokollen
  min_dbm_history[history_idx] = measured_min_dbm;
  max_dbm_history[history_idx++] = measured_max_dbm;  
  if (history_idx >= MAX_HISTORY_COUNT)
    history_idx = 0;                                          //Reset ringbuffer Zeiger
  
  // Holen des absoluten Min/Max-Wertes aus dem Verlaufsprotokollen
  min_dbm = 1000, max_dbm = -1000;
  for (int h=0; h < MAX_HISTORY_COUNT; h++)
  {
    if (min_dbm_history[h] < min_dbm)
      min_dbm = min_dbm_history[h];                           // Leistung minimum Wert
    if (max_dbm_history[h] > max_dbm)
      max_dbm = max_dbm_history[h];                           // Leistung maximum Wert
  }    
  // Berechnen Sie die Modulationsindex-basierte Spannung von min und max dbm-Werten.
  float min_voltage, max_voltage;
  max_voltage = dBmToVolt(measured_max_dbm);
  min_voltage = dBmToVolt(measured_min_dbm);   
  modulation_index = ((max_voltage - min_voltage) / (max_voltage + min_voltage) * 100.0);   
}

//******************************** Convertiert die AD-Wandler Werte in Spannung und in dbm ********
float measuredValueTodBm(int value) 
{
  float voltage = value * (V_REF / AD_AUFLOESUNG);    
  float dbm = convertVoltageToDbm(voltage);
  return dbm;
}

// **********************  Funktion Reset den Merkerpuffer ***************************************************************************************************
void resetHistoryRingBuffers() 
{
  for (int h=0; h < MAX_HISTORY_COUNT; h++)
  {
    min_dbm_history[h] = 1000;      
    max_dbm_history[h] = -1000;      
  }  
}
//***************************************************************************************************************************************************************
// * Überprüft, ob die Leitungsmessung innerhalb der vorgegebenen Grenzen ist
// * Overlad am Pegelmess-IC kann es sehr leicht zerstören
//****************************************************************************************************************************************************************
bool checkInputBounds() 
{
  power_overload = (avg_dbm >= ABS_MAX_INPUT_DBM);
  power_notconnected = (avg_dbm < MIN_INPUT_DBM);  
  return !power_overload;
}

//************************************************* Funktion, löscht ausgewählte Kalibrierpunkte*******************************************************************
void deleteCalibrationPoint(byte pointToDelete) {
  //*********************************************** Daten des zu löschenden Punktes zurücksetzen. Frequenz von "0" makiert als gelöscht.
  calibrationData[pointToDelete].frequenz = 0;
  calibrationData[pointToDelete].dBm_at_0V = 0;  
  selected_calibration_point--; //****************** Vorherigen Punkt oder Menü auf lcd anzeigen

  // Höhere Frequenzen im Array um eins nach unten verschieben
  if (pointToDelete < (MAX_CALIBRIERUNGS_PUNKTE - 1))
  {
    for (int i=pointToDelete; i < (MAX_CALIBRIERUNGS_PUNKTE - 1); i++)
    {
      calibrationData[i].frequenz = calibrationData[i+1].frequenz;
      calibrationData[i].dBm_at_0V = calibrationData[i+1].dBm_at_0V;  
      calibrationData[i+1].frequenz = 0;
      calibrationData[i+1].dBm_at_0V = 0;  
    }
  }
  writeCalibrationDataToEEPROM();   // Write to EEPROM
  mergeCalibrationData();           // Recalculate merged data
}

// ************************Funtion,  zur nächsten Kalibrieroption navigieren ********************************
void nextCalibrationPoint(){  
  selected_calibration_point++;  
  //************************ Prüfen Sie, ob wir den letzten kalibrierten Punkt erreicht haben. **************
  if (selected_calibration_point >= MAX_CALIBRIERUNGS_PUNKTE || (selected_calibration_point >= 0 && calibrationData[selected_calibration_point].frequenz == 0))
    selected_calibration_point = -2; // Show exit screen     
}

//************************ Funktion, Zur vorherigen Kalibrieroption navigieren ******************************
void previousCalibrationPoint(){
  // Goto Previous calibration point. If not there then show 'Add new calibration'
  selected_calibration_point--;
  if (selected_calibration_point < -2)
  { 
    // ******************  Ermitteln des letzten Kalibrierpunktes im Array ***********************
    for (int i = MAX_CALIBRIERUNGS_PUNKTE-1; i>=-1; i--)
    {
      if (i==-1)
      {
        // *************  Keine Kalibrierpunkte im Array *****************************************
        selected_calibration_point = -1; // Show 'Add new calibration'
        break;
      }
      else if (calibrationData[i].frequenz > 0)
      {
        selected_calibration_point = i;  //**** Letzter Kalibrierpunkt im Array *****************
        break;
      }
    }            
  }  
}

// ********************** Funktion, Prüfen, ob Platz für einen anderen Kalibrierpunkt vorhanden ist. *****************************************
bool calibrationSlotLeft()
{
  for (int i=0; i < MAX_CALIBRIERUNGS_PUNKTE; i++)
  {
    if (calibrationData[i].frequenz == 0)
      return true;    
  }
  return false;
}

// ********************************Funktion, Sichern der eingegebenen Kalibrierdaten und Speichern im EEPROM  *******
void saveCalibrationData()
{
  int i=0;
  float dbm_at_0V;
  float measuredVoltage = avg_volt; // Aktuell gemessene avg_volt abfragen
  // ************    Wandelt die gemessene Leistung in einen dbm-Wert bei 0V um ****************************
  // ************    Wenn die gemessene Spannung 700mV bei -10dBm beträgt. Dann mit einer Steigung von -22mV/dBm ==> (700/22m) - 10 = 21.82dBm bei 0V.
  dbm_at_0V = (measuredVoltage / MV_DB_SLOPE) + calibrated_power;

  // ******************************* Daten in leeren Speicherplatz legen **************************************************************
  while (i < MAX_CALIBRIERUNGS_PUNKTE)
  {
    if (calibrationData[i].frequenz == 0)
    {
      // Empty slot
      calibrationData[i].frequenz = menu_freq;
      calibrationData[i].dBm_at_0V = dbm_at_0V;
      selected_calibration_point = i;
      break;
    }
    i++;
  }
  // Save updated data to EEPROM
  writeCalibrationDataToEEPROM();
}
// ***************************** Prüfen Sie, ob die eingegebene Kalibrierfrequenz bereits in der Liste enthalten ist.
// ***************************** Es sind keine Duplikate erlaubt.
bool calibrationExists(int frequenz)
{
  for (int i=0; i < MAX_CALIBRIERUNGS_PUNKTE; i++)
  {
    if (calibrationData[i].frequenz == frequenz)
      return true;    
  }
  return false;
}

//**************************** Ausgabe der Leistung  in uW, mW or W. ******************************************
byte printPowerWatts(float mw) 
{
  byte chars = 0;

   if (gross == 1){                                         // Große Schrift.
     display.setTextSize(2);
     display.setCursor(0, 50);
     gross = 0;
   }     
    else
    {
     display.setTextSize(1);
     display.setCursor(64, 20);
    }

//######################################## Anzeige Leistung mittig ! ############################################################################

  if (mw < 0.001)                                                                     //********* Wenn weniger als 10uW Anzeige in nW 
  {
    watt = 1;
    chars += printFormattedNumber(mw * 1000000.0, 3, 0,  false, false);
    display.print(F(" nW"));
  }
  else if (mw < 0.01)                                                                 //***** Wenn weniger als 10uW Anzeige in 0,1 uW Auflösung
  {
    chars += printFormattedNumber(mw * 1000.0, 3, 0,  false, false);
    display.print(F(" uW"));
  }
  else if (mw < 1.0)                                                                  //****** Bei weniger als 1mW Anzeige in uW
  {
    uw = 1;
    chars += printFormattedNumber(mw * 1000.0, 3, 0,  false, false);
    display.print(F(" uW"));
    uw = 0;
  }
  else if (mw >= 1000.0)                                                              //*** Wenn mehr als 1000mW Anzeige in W
  {
    chars += printFormattedNumber(mw / 1000.0, 3, 1, false, false);
    display.print(F(" W"));
  }
  else
  {
    chars += printFormattedNumber(mw, 3, 1,  false, false);
    display.print(F(" mW"));
  }
  watt = 0;
  display.display();
  return chars;
}

// ***********************************Funktion, nächstes oder vorheriges Vordämpfungeinstellung wählen **********************
void setAttenuationProfile(bool next)
{
  if (next && att_profile < MAX_ATTENUATION_PROFILE)
    att_profile++;  
  else if (next && att_profile >= MAX_ATTENUATION_PROFILE)
    att_profile = 0;
  else if (!next && att_profile > 0)
    att_profile--;
  else if (!next && att_profile <= 0)
    att_profile = MAX_ATTENUATION_PROFILE;
}

// ******************************** Funktion, füllen Sie eine komplette Displayzeile mit Leerzeichen aus. *******************####################
//void filloutLine(byte charsWritten) 
//{
//  for (byte i=0; i < 16 - charsWritten;i++)
//      display.print(" ");   
//}
//****************************************************************************************************************
/* Gibt eine Zahl mit einer bestimmten Anzahl von Ziffern und Dezimalstellen auf dem Display aus.
 * Die Funktion rundet die Zahl nicht ab.
 * Es ist auch möglich, ein Leerzeichen zu drucken, wenn es kein negatives Vorzeichen gibt..
 * Ziffer: Ziffern, die angezeigt werden soll
 * minAmountDigits: Die minimale Anzahl der anzuzeigenden Ziffern. Wenn die Zahl kürzer ist, werden vor der Zahl 0's gedruckt.
 * amountDecimals: Die Anzahl der anzuzeigenden Dezimalstellen.
 * fillout:Wenn die Anzahl der Dezimalstellen kürzer als die gewünschte Anzahl ist, werden Leerzeichen verwendet, um die gleiche Stringlänge zu erhalten.
*/
byte printFormattedNumber(float number, byte minAmountDigits, byte amountDecimals, bool reserveMinus, bool fillout)
{
  byte chars = 0; // ***************  Anzahl der geschriebenen Zeichen...******************************************
  
  // ******************************* Wenn kein Minuszeichen, schreiben Sie ein Leerzeichen.************************
  if (reserveMinus && number >= 0){
    chars += display.print(" ");
    dbm = 1;}
  else if (number < 0) {
    chars += display.print("-");
    dbm = 1;
    number *= -1;
  }
    
  //****************************** Wenn der aktuelle Wert weniger Ziffern hat, dann werden zusätzliche Nullen hinzugefügt.
  int digitPart = (int)number;
  byte digitCount = getAmountDigits(digitPart);

  if ((int)number >= 100 && uw == 1){display.print("  ");}                       //################################################
  
  if (digitCount < minAmountDigits)
  {
  if (amountDecimals == 0 && kallibrieren_nicht == 1) {minAmountDigits = (minAmountDigits + 2); }
    for (byte i=0; i < minAmountDigits - digitCount; i++)

      if (dbm == 1){          
        chars += display.print(" "); 
        dbm = 0;}
      else {
        if (watt == 0){
          chars += display.print("0"); }
        else {
          chars += display.print(" "); }
        }
  }
  chars += display.print(number, amountDecimals); 
 
 // ************************** Fehlende Zeichen nach gewünschter Länge ausfüllen ****************************************
  if (fillout) 
  {
    byte totalCharsWanted = minAmountDigits + amountDecimals;
    if (reserveMinus)
      totalCharsWanted++;    
 
    for (byte i=0; i < totalCharsWanted - chars; i++)
       chars += display.print("0");
  }  
  return chars; // ************** gibt die Anzahl der Zeichen zurück, die in das Display geschrieben wurden. **************
}

// ***************************** Anzahl der Stellen einer Ganzzahl ermitteln **********************************************.
// ***************************** Einfach aber schnell *********************************************************************
byte getAmountDigits(int n)
{
    n = abs(n);
    if (n < 10) return 1;
    if (n < 100) return 2;
    if (n < 1000) return 3;
    if (n < 10000) return 4;    
}

// ************************* Ermittelt die aktuell gewählte Vordämpfung ****************************************************

 float calculateAttenuation()
{
  float att_value;
  switch (att_profile)
  {
    case 0: //************* Benutzerdefinierter Festwert *************************************
      att_value = custom_attenuation;
      break;
    case 1: //************* VAT-20W2+, wählen Sie die Dämpfung der aktuell gewählten Frequenz aus. *************************
      att_value = interpolatedAttenuatorValue(VAT20W2_ATT_PROFILE);
      break;
    case 2: //************* VAT-30W2+, wählen Sie die Dämpfung der aktuell gewählten Frequenz aus. *************************
      att_value = interpolatedAttenuatorValue(VAT30W2_ATT_PROFILE);
      break;
  }
  return att_value;
}
 
// ********************** Interpolierte Dämpfung aus Array-Werten ********************************************************** 
float interpolatedDbmValue()
{  
  int minIdx = 0;
  int amountItems = MAX_CALIBRIERUNGS_PUNKTE + ANZ_CAL_FREQUENZEN;
  for (minIdx; minIdx < amountItems; minIdx++)
  {
      if (selected_freq == mergedCalibrationData[minIdx].frequenz) //********* Wenn exakter Wert.**************************
        break;           
      else if (selected_freq < mergedCalibrationData[minIdx].frequenz || mergedCalibrationData[minIdx].frequenz == 0) {
        minIdx--;   //********** Einer zu weit *****************************************************************************
        break;
      }
  }  

  if (mergedCalibrationData[minIdx].frequenz == selected_freq || ((minIdx + 1) == amountItems) ) //*******   Exakter Wert oder letzter Wert aus der Liste
    return mergedCalibrationData[minIdx].dBm_at_0V;    //**************************************** Nichts zu interpolieren, nur Wert nehmen.***************
  
  // Interpolate array
  int minFreq = mergedCalibrationData[minIdx].frequenz;
  int maxFreq = mergedCalibrationData[minIdx + 1].frequenz;
  if (maxFreq == 0)
  {
    //****************************** Es gibt keinen höheren Datenpunkt. Also nichts zu interpolieren. Rückgabe unterer Datenpunktwert *********************
    return mergedCalibrationData[minIdx].dBm_at_0V;
  }
  float minValue = mergedCalibrationData[minIdx].dBm_at_0V;
  float maxValue = mergedCalibrationData[minIdx + 1].dBm_at_0V;
  int freq_range = maxFreq - minFreq;
  int freq_diff = selected_freq - minFreq;
  float ratio = ((float)freq_diff) / ((float)freq_range);

  float value_range = maxValue - minValue;
  return (value_range * ratio) + minValue;  
}
//*********************************** Interpolierte Dämpfung aus den gelieferten Array-Werten ************************************************************
float interpolatedAttenuatorValue(const float* values)
{
  int minIdx = 0;
  for (minIdx; minIdx < ANZ_CAL_FREQUENZEN; minIdx++)
  {
      if (selected_freq == FREQUENZEN[minIdx])
        break;           
      else if (selected_freq < FREQUENZEN[minIdx]) {
        minIdx--;
        break;
      }
  }  

  if (FREQUENZEN[minIdx] == selected_freq || ((minIdx + 1) == ANZ_CAL_FREQUENZEN)) 
    return values[minIdx];    //*************************** Nichts zu interpolieren ****************************************
  
  //******************************** Interpolier  Array ********************************************************************
  int minFreq = FREQUENZEN[minIdx];
  int maxFreq = FREQUENZEN[minIdx + 1];
  float minAtt = values[minIdx];
  float maxAtt = values[minIdx + 1];
  int freq_range = maxFreq - minFreq;
  int freq_diff = selected_freq - minFreq;
  float ratio = ((float)freq_diff) / ((float)freq_range);

  float att_range = maxAtt - minAtt;
  return (att_range * ratio) + minAtt;  
}
  
//****************************** Berechnet Sie den aktuellen Messfehler. ****************************************************
// Der Fehler hängt stark von der Temperatur ab. Also machen wir es auf die einfache Art und Weise.
// und nehmen Sie den nächsten durchschnittlichen Fehler.
// **************************************************************************************************************************
float currentError() 
{
  int roundedDbm = (int)(round(avg_dbm / 5.0) * 5.0);   //****************** Rund bis auf 5 dB genau ************************
  int idx = (roundedDbm + 60) / 5;                      //****************** Liefert den Array-Index. 0 = -60dBm ************
  if (idx < 0)                                          //****************** Scheckgrenzen **********************************
    idx = 0;
  else if (idx >= (sizeof(ERROR_DBM) / sizeof(float))) 
    idx = (sizeof(ERROR_DBM) / sizeof(float)) - 1; 
  return ERROR_DBM[idx];      
}

//****************************************************Funktion, Initialisierung der Eingabetastenabfrage ****************************
void initButtons() 
{
  // Set button pins as input.  
  pinMode(TASTE_AUF ,INPUT_PULLUP);                     // PullUp einschalten #####################################################
  pinMode(TASTE_EING ,INPUT_PULLUP);
  pinMode(TASTE_AB ,INPUT_PULLUP);

  TASTE_AUF , !TASTE_AUF;                               //Tasten invertieren um die Widerstände zu sparen. ########################
  TASTE_EING , !TASTE_EING;                             //IM Prog. Menü sind auch noch sachen geändert worden. ####################
  TASTE_AB , !TASTE_AB;
  
  // Attach pins to Bounce
  button_UP.attach(TASTE_AUF);
  button_ENTER.attach(TASTE_EING);
  button_DOWN.attach(TASTE_AB);

  // Set bounce interval
  button_UP.interval(TASTE_ENTPRELLEN_MS);
  button_ENTER.interval(TASTE_ENTPRELLEN_MS);
  button_DOWN.interval(TASTE_ENTPRELLEN_MS);
}

// **************************************** Funktion, fügt calibrationData und DBM_AT_0V zu einem Array vom Typ calibrationPoint zusammen und sortiert nach Häufigkeit. 
void mergeCalibrationData()
{
  // Clear current
  for (int i=0; i<MAX_CALIBRIERUNGS_PUNKTE + ANZ_CAL_FREQUENZEN; i++)
  {
    mergedCalibrationData[i].frequenz = 0;
    mergedCalibrationData[i].dBm_at_0V = 0;
  }
  
  //************************************* Füllt Sie das zusammengefügte Array mit den Kalibrierdaten des Benutzers. **************************
  for (int i=0; i<MAX_CALIBRIERUNGS_PUNKTE; i++)
  {    
    mergedCalibrationData[i].frequenz =  calibrationData[i].frequenz;
    mergedCalibrationData[i].dBm_at_0V =  calibrationData[i].dBm_at_0V;    
  }
  
  //************************************ Fügt die Original Kalibrierdaten hinzu. Aber fügt nichts hinzu, wenn die Frequenz bereits vorhanden ist.
  for (int i=0; i<ANZ_CAL_FREQUENZEN; i++)
  {
    //********************************** Überprüfen Sie, ob bereits im Array vorhanden ist. **************************************************
    bool alreadyExists = false;
    for (int j=0; j<MAX_CALIBRIERUNGS_PUNKTE; j++)
    {
      if (mergedCalibrationData[j].frequenz == FREQUENZEN[i])
      {
        alreadyExists = true;
        break;
      }
    }   
    if (!alreadyExists)
    {    
      //******************************** Wenn es nicht existiert, fügen Sie dem Array hinzu.******************************************************
      mergedCalibrationData[i + MAX_CALIBRIERUNGS_PUNKTE].frequenz = FREQUENZEN[i];
      mergedCalibrationData[i + MAX_CALIBRIERUNGS_PUNKTE].dBm_at_0V = DBM_AT_0V[i];
    }
  }

  //*********************************** Wir wollen, dass Elemente mit der Häufigkeit=0 nach dem Sortieren am Ende des Arrays stehen.************************
  // ********************************** Also ändern wir sie vorübergehend auf eine Zahl, die größer ist als die größte Frequenz, die wir eingeben können.***
  for (int i=0; i<MAX_CALIBRIERUNGS_PUNKTE + ANZ_CAL_FREQUENZEN; i++)
  {
    if (mergedCalibrationData[i].frequenz == 0)
      mergedCalibrationData[i].frequenz = 9999;
  }

  // ******************************** Sortiert nach Frequenz ****************************************
  bubleSort(mergedCalibrationData, MAX_CALIBRIERUNGS_PUNKTE + ANZ_CAL_FREQUENZEN);

  //********************************* Wiederherstellen der Frequenz auf 0 ***************************
  for (int i=0; i<MAX_CALIBRIERUNGS_PUNKTE + ANZ_CAL_FREQUENZEN; i++)
  {
    if (mergedCalibrationData[i].frequenz == 9999)
      mergedCalibrationData[i].frequenz = 0;
  }

  // Debug stuff
}

//*********************************************Funktion,  Buble-Sortierfunktion zum Sortieren von Frequenzen in der KalibrierungData ****************************
void bubleSort(calibrationPoint *data, int size) 
{
  for(int i=0; i<(size-1); i++) {
    for(int o=0; o<(size-(i+1)); o++) {
      if(data[o].frequenz > data[o+1].frequenz) {
        calibrationPoint t = data[o];
        data[o] = data[o+1];
        data[o+1] = t;        
      }
    }
  }
}

//*********************************************Funktion  Lese die einstellungen aus dem  EEPROM *****************************************************************
void getSettingsFromEEPROM() 
{ 
  //******************************************* Prüfen Sie, ob das EEPROM initialisiert wurde.
  short schema = 0;
  EEPROM_readAnything(INIT_EEPROM_ADDRESS, schema);  
  if (schema != INIT_SCHEMA)   // ************* Vergleich mit erwartetem Schema   ***************
    initEEPROM();   // Initialize EEPROM with default value

  //***************************** Lese Daten vom EEPROM
  EEPROM_readAnything(FREQ_EEPROM_ADDRESS, selected_freq);  
  EEPROM_readAnything(ATTENUATION_PROFILE_EEPROM_ADDRESS, att_profile);  
  EEPROM_readAnything(ATTENUATION_VALUE_EEPROM_ADDRESS, custom_attenuation);  
  getCalibrationDataFromEEPROM();
}

void getCalibrationDataFromEEPROM()
{
  EEPROM_readAnything(CALIBRATION_DATA_EEPROM_ADDRESS, calibrationData);   
}

// ****************************** Funktion, Initialisiere das EEPROM mit  Vorgabewerten *******************
void initEEPROM()
{
  //**************************** Setze standart Werte *******************************************
  selected_freq = FREQ_DEFAULT;
  custom_attenuation = ATTENUATION_DEFAULT;
  att_profile = ATTENUATION_PROFILE_DEFAULT;  
  writeSettingsToEEPROM();
  
  //*************************** Lösche die  Calibrierung ****************************************
  for (int i=0; i < MAX_CALIBRIERUNGS_PUNKTE; i++)
  {
    calibrationData[i].frequenz = 0;
    calibrationData[i].dBm_at_0V = 0;
  }  
  writeCalibrationDataToEEPROM();
  
  //************************** Schreiben Sie ein Schema, damit wir diese Funktion beim nächsten Mal überspringen können.
  EEPROM_writeAnything(INIT_EEPROM_ADDRESS, INIT_SCHEMA); // Mark EEPROM as initialized
}

// **************************** Schreibe Einstellungen ins EEPROM *********************************
void writeSettingsToEEPROM() 
{
  EEPROM_writeAnything(FREQ_EEPROM_ADDRESS, selected_freq);  //****************** Schreibe gewählte Frequenz ins EEPROM  
  EEPROM_writeAnything(ATTENUATION_PROFILE_EEPROM_ADDRESS, att_profile);
  EEPROM_writeAnything(ATTENUATION_VALUE_EEPROM_ADDRESS, custom_attenuation);  // ** Schreibe gewählten abschwächer ins EEPROM  
}

// ********************************************************* Funktion  Schreib Kalibrierdatn ins EEPROM ****************************************************
void writeCalibrationDataToEEPROM()
{
  EEPROM_writeAnything(CALIBRATION_DATA_EEPROM_ADDRESS, calibrationData);
}

//***************************************************************************************************
// ************** Wandelt Spannung in dBm.
// ************** Gemäß Abbildung 8 des Datenblatts
// ************** dBm value at 0V - (measured voltage / slope)
// ************** Der unterstützte Bereich ist -50dBm bis 0dBm wegen des hohen Fehlers darüber hinaus. 
float convertVoltageToDbm(float voltage) 
{
  return dbm_at_0V - (voltage / MV_DB_SLOPE);
}

//*************** Convertiert dBm to mW *********
float convertDbmToMilliWatt(float dBm) 
{  
  return pow(10, dBm / 10.0);
}

//************** Convertiert mW to dBm **********
float convertMilliWattToDbm(float mw) 
{
  return 10 * log10(mw);
}

//************** Convertiert dBm to voltage ****
float dBmToVolt(float dBm) 
{
  return pow(10, dBm/20.0);
}

//************** Auf eine bestimmte Anzahl von Dezimalstellen runden ********
float roundDecimals(float value, int decimals)
{
  float d = pow(10, decimals);
  if (value < 0)
    return ceil((value * d - 0.5)) / d;
  else 
    return floor((value * d + 0.5)) / d;
}

// ***************  Speichern beliebiger Typen im EEPROM  ********************
template <class T> int EEPROM_writeAnything(int adr, const T& value)
{
    const byte* p = (const byte*)(const void*)&value;
    unsigned int i;
    for (i = 0; i < sizeof(value); i++)
          EEPROM.write(adr++, *p++);
    return i;
}

// *************** Lesen beliebiger Typen aus dem  EEPROM ********************
template <class T> int EEPROM_readAnything(int adr, T& value)
{
    byte* p = (byte*)(void*)&value;
    unsigned int i;
    for (i = 0; i < sizeof(value); i++)
          *p++ = EEPROM.read(adr++);
    return i;
}

//*********************************************************   Programmende Stand 03.01.2018   ******************************************************
